<div class="rst-versions" data-toggle="rst-versions" role="note" aria-label="versions">
  <span class="rst-current-version" data-toggle="rst-current-version">
    Version: {{ current_version }}
    <span class="fa fa-caret-down"></span>
  </span>
  <div class="rst-other-versions">
    {% if languages|length >= 1 %}
    <dl>
      <dt>{{ _('Languages') }}</dt>
        {% for the_language, url in languages %}
          <dd><a href="{{ url }}/index.html">{{ the_language }}</a></dd>
        {% endfor %}
    </dl>
    {% endif %}
    {% if versions|length >= 1 %}
    <dl>
      <dt>{{ _('Versions') }}</dt>
      {% for the_version, url in versions %}
        <dd><a href="{{ url }}/index.html">{{ the_version }}</a></dd>
      {% endfor %}
    </dl>
    {% endif %}
    <br>
    </dl>
  </div>
</div>
